import numpy as np
import pandas as pd

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import StandardScaler, PowerTransformer
from sklearn.decomposition import PCA

from rdkit import rdBase, Chem, DataStructs
from rdkit.Chem import AllChem, Descriptors, Draw, MolStandardize
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.ML.Descriptors import MoleculeDescriptors

from mordred import Calculator, descriptors
from umap import UMAP

from bokeh.plotting import figure, show
from bokeh.models import HoverTool, LinearColorMapper, LogColorMapper, ColorBar
from bokeh.transform import transform
from bokeh.io import push_notebook, output_notebook, push_notebook, save
from bokeh.palettes import Spectral, Reds256, Blues256
from ipywidgets import interact


def filter_columns(df, columns):
    """
    指定された列のみを選択してデータフレームを返す関数です。    
    """
    return df.loc[:, columns]
    

def remove_columns(df, columns):
    """
    指定された列を削除する関数です。存在しない列は無視されます。
    """
    return df.drop(columns, axis=1, errors="ignore")


def standardize_moelcule(df, molCol="ROMol"):
    """
    複数のフラグメントの中から最も大きなフラグメントを選択します。
    """
    # ログを抑制
    rdBase.DisableLog("rdApp.info")
    # LargestFragmentChooserのインスタンスを作成
    lfc = MolStandardize.rdMolStandardize.LargestFragmentChooser()
    # 各分子について最も大きなフラグメントを選択
    df[molCol] = df[molCol].apply(lfc.choose)
    uc = MolStandardize.rdMolStandardize.Uncharger()
    df[molCol] = df[molCol].apply(uc.uncharge)
    # ログ抑制を戻す
    rdBase.EnableLog("rdApp.info")
    return df

    
def calc_rdkit_descriptor(df, molCol="ROMol"):
    """
    RDKit記述子を計算してデータフレームに追加します。
    """
    
    # 警告を抑制
    rdBase.DisableLog("rdApp.warning")
    # RDKitの記述子の名前を取得
    desc_names = [desc[0] for desc in Descriptors.descList]
    # 記述子計算機を初期化
    descriptor_calculator = MoleculeDescriptors.MolecularDescriptorCalculator(desc_names)
    # 各分子に対して記述子を計算
    descriptors = [descriptor_calculator.CalcDescriptors(mol) for mol in df[molCol].values]
    # 計算した記述子をデータフレームに変換        
    df_descriptors = pd.DataFrame(descriptors, columns=desc_names, index=df.index)
    # マージ
    df = df.join(df_descriptors)
    # 警告抑制を戻す
    rdBase.EnableLog("rdApp.warning")
    return df        


def calc_mordred_descriptor(df, molCol="ROMol", ignore_3D=False):    
    """
    Mordred記述子を計算してデータフレームに追加します。
    """
    # Mordred記述子計算器を初期化
    mordred_calculator = Calculator(descriptors, ignore_3D=ignore_3D)
    # 指定された列の分子オブジェクトから記述子を計算
    df_descriptors = mordred_calculator.pandas(df[molCol])
    # 数値型に変換。計算エラーが文字列として出力された場合はnanに変換
    df_descriptors = df_descriptors.apply(pd.to_numeric, errors='coerce')
    # 元データと同名の列があれば削除する処理
    for column in df_descriptors.columns:
        if column in df.columns:
            df = df.drop(column, axis=1)
    # 元データと計算された記述子を結合
    df = df.join(df_descriptors)

    return df

        
class NullColumnFilter(TransformerMixin, BaseEstimator):
    """
    欠損値の割合が指定した閾値（基準値）を超える列を削除するクラスです。
    """

    def __init__(self, threshold=0.5):
        # 閾値を設定
        self.threshold = threshold
        
    def fit(self, X, y=None):
        # 欠損値の割合が閾値を超えない列を選択
        self.valid_columns = list(X.columns[~(X.isnull().mean() > self.threshold)])
        return self
        
    def transform(self, X):
        # フィルタリングされた列のみを保持
        return X.loc[:, self.valid_columns]


class CorrelationFilter(TransformerMixin, BaseEstimator):
    """
    相関係数が指定した閾値以上の列の片方を削除するクラスです。
    """

    def __init__(self, threshold=0.95):
        self.threshold = threshold
        self.valid_columns = None

    def fit(self, X, y=None):
        df_corr = X.corr()

        while True:    
            # 現在の相関行列のカラム名を取得
            columns = df_corr.columns
            cols_to_drop = set()
            cols_associated_drop_column = set()
                
            for i, query in enumerate(columns):
                if query in cols_to_drop:
                    continue
                
                # 自分を除くシリーズを作成
                sr = df_corr[query].drop(index=query)
        
                # 自身と相関が閾値以上の特徴を取得
                sr = sr[sr.abs() >= self.threshold]        
                
                for target, value in sr.items():
                    # 特徴が削除根拠になっている場合は削除不可
                    if target in cols_associated_drop_column:
                        continue
                        
                    # 特徴が既に削除リストにある場合は対象外
                    if target in cols_to_drop:
                        continue            
                        
                    # 特徴を削除対象に追加
                    cols_to_drop.add(target)
                    
                    # 自身は削除根拠リストに追加
                    cols_associated_drop_column.add(query)
                    print(f"{target} is dropped because correlation with {query} is {value}")
        
            # 削除対象の列がない場合は終了
            if len(cols_to_drop) == 0:
                break
                
            # 相関行列の更新
            df_corr = df_corr.drop(columns=cols_to_drop, index=cols_to_drop)
        
        self.valid_columns = list(df_corr.columns.values)
        return self

    def transform(self, X):
        if self.valid_columns:
            return X.loc[:, self.valid_columns]
        else:
            return X


class KNNBasedReliabilityTransformer(TransformerMixin, BaseEstimator):
    """
    k近傍法を用いてデータの信頼性を評価するクラス。
    各データ点のk近傍の距離の平均を計算し、閾値を設定して異常値を判定します。
    """
    
    def __init__(self, feature_columns, k=5, metric='euclidean'):
        self.feature_columns = feature_columns
        self.k = k
        self.metric = metric
        self.transformer = None
        
    def fit(self, X, y=None):
        # 特徴カラムを選択
        X_features = X[self.feature_columns]
        
        # k近傍法モデルを初期化し、学習データでフィッティング
        self.model = NearestNeighbors(n_neighbors=self.k, metric=self.metric)
        self.model.fit(X_features)

        # 学習データにおける他学習データとのK+1近傍の距離を取得(自身も含まれるためK+1としている))
        knn_distance, _ = self.model.kneighbors(X_features, n_neighbors=(self.k + 1))

        # 自身を除いたK近傍の距離の平均を計算
        mean_of_knn = np.mean(knn_distance[:, 1:], axis=1)

        # 距離の平均の分布を正規分布に変換
        self.transformer = PowerTransformer(method='yeo-johnson')
        mean_of_knn = mean_of_knn.reshape(-1, 1)
        transformed_data = self.transformer.fit_transform(mean_of_knn).flatten()

        # 信頼性=1とみなす最小値と、信頼性0とみなす3シグマ値に対応する値を計算
        self.min_value = np.min(transformed_data)
        mean_value = np.mean(transformed_data)
        std_value = np.std(transformed_data)
        self.three_sigma = mean_value + 3 * std_value        
        
        return self

    # 信頼性の計算関数
    def reliability(self, x, min_val, three_sigma):
        if x <= min_val:
            return 1
        elif min_val < x < three_sigma:
            return (three_sigma - x) / (three_sigma - min_val)
        else:
            return 0            
        
    def transform(self, X):
        # 特徴カラムを選択
        X_features = X[self.feature_columns]
        
        # 予測データのK近傍にある学習データとの距離を取得
        knn_distance_predict, _ = self.model.kneighbors(X_features, n_neighbors=self.k)
        
        # 距離の平均を算出
        mean_of_knn_predict = np.mean(knn_distance_predict, axis=1)
        X["distance"] = mean_of_knn_predict

        # 距離の値が正規分布のどこに位置するかを計算
        mean_of_knn_predict = mean_of_knn_predict.reshape(-1, 1)
        transformed_data_predict = self.transformer.transform(mean_of_knn_predict).flatten()

        # 信頼性の値を計算
        X["reliability"] = [self.reliability(x, self.min_value, self.three_sigma) for x in transformed_data_predict]
        
        return X


class UMAPTransformer(TransformerMixin, BaseEstimator):    
    """
    UMAPを使用してデータの次元を削減し、低次元空間に変換する
    """
    
    def __init__(self, feature_columns, n_components=2, random_state=42):
        self.feature_columns = feature_columns
        self.n_components = n_components
        self.random_state = random_state
        
    def fit(self, X, y=None):
        # 特徴カラムを選択
        X_features = X[self.feature_columns]
        
        # UMAPモデルの初期化と適用
        self.model = UMAP(n_components=self.n_components, random_state=self.random_state)
        self.model.fit(X_features)
        
        return self
        
    def transform(self, X):
        X_features = X[self.feature_columns]
        embedding = self.model.transform(X_features)
        
        # 変換後のデータフレームにUMAPの座標を追加
        X_umap = X.copy()
        X_umap[[f"umap_{i+1}" for i in range(self.n_components)]] = embedding
        return X_umap


def smiles2svg(smiles, width=200, height=100):
    """
    SMILES 文字列から分子構造を描画し、SVG形式で返す。
    """
    # SMILES文字列から分子オブジェクトを作成
    mol = Chem.MolFromSmiles(smiles)
    
    # 描画用のSVGオブジェクトを初期化
    drawer = rdMolDraw2D.MolDraw2DSVG(width, height)
    
    # 分子を描画
    drawer.DrawMolecule(mol)
    
    # 描画の終了とSVGテキストの取得
    drawer.FinishDrawing()
    return drawer.GetDrawingText()


def plot_data(df_train, df_predict=None):
    """
    学習データ、予測データ(オプション)を可視化するための散布図を生成します。
    データの色付けは、ターゲットの値または信頼性に基づいて行われ、これら切り替えるためのセレクトボックスが表示されます。
    """

    # Bokehの出力設定
    output_notebook()    
    
    # ツールチップの共通部分
    def create_tooltip(name_label, value_label, reliability_label, value_label_text):
        return f"""
        <div>
        <b>name</b>: @{name_label}<br>
        <b>{value_label_text}</b>: @{value_label}{{0.00000}}<br>
        <b>reliability</b>: @reliability{{0.00000}}<br>
        <div>@image{{safe}}</div>
        </div>
        """

    # カラーマッパーを作成する関数(学習データ・テストデータ)
    def create_color_mapper_train_pred(data1, data2, palette):
        tmp_min = min(data1.min(), data2.min())
        tmp_max = max(data1.max(), data2.max())
        return LinearColorMapper(palette=palette, low=tmp_min, high=tmp_max)

    # カラーマッパーを作成する関数(学習データ)
    def create_color_mapper_train(data, palette):
        tmp_min = data.min()
        tmp_max = data.max()
        return LinearColorMapper(palette=palette, low=tmp_min, high=tmp_max)
    
    # カラーマッパーの設定
    mappers = {}    
    if df_predict is not None:    
        mappers = {
            "target": create_color_mapper_train_pred(df_train["target"], df_predict["prediction_label"], Reds256[::-1]),
            "reliability": create_color_mapper_train_pred(df_train["reliability"], df_predict["reliability"], Blues256[::-1])
        }
        plot = figure(width=800, height=500, title='Distribution of Training and Prediction Data', 
              x_axis_label='umap_1', y_axis_label='umap_2')
    else:
        mappers = {
            "target": create_color_mapper_train(df_train["target"], Reds256[::-1]),
            "reliability": create_color_mapper_train(df_train["reliability"], Blues256[::-1])
        }        
        plot = figure(width=800, height=500, title='Distribution of Training Data', 
              x_axis_label='umap_1', y_axis_label='umap_2')        
        
    # 色付けの対象（ターゲットの値、信頼性による色付け)
    color_target="target" 
    
    # 学習データのプロット
    r_train = plot.scatter(
        'umap_1', 'umap_2', size=10, source=df_train, 
        line_color="black", alpha=0.8, legend_label="train", 
        color=transform("target", mappers[color_target]), marker="triangle"
    )
    
    # ツールチップの定義（学習データ用）
    TOOLTIPS_TRAIN = create_tooltip("ID", "target", "reliability", "exp")
    hover_tool_train = HoverTool(tooltips=TOOLTIPS_TRAIN)
    hover_tool_train.renderers = [r_train]
    plot.add_tools(hover_tool_train)

    if df_predict is not None:
        # 予測データのプロット
        r_pred = plot.scatter(
            'umap_1', 'umap_2', size=10, source=df_predict, 
            line_color="black", alpha=0.8, legend_label="predict", 
            color=transform("prediction_label", mappers[color_target]), marker="circle"
        )

        # ツールチップの定義（予測データ用）
        TOOLTIPS_PRED = create_tooltip("ID", "prediction_label", "reliability", "prediction")
        hover_tool_pred = HoverTool(tooltips=TOOLTIPS_PRED)
        hover_tool_pred.renderers = [r_pred]
        plot.add_tools(hover_tool_pred)
    
    
    # 凡例のクリックポリシー設定（クリックで非表示）
    plot.legend.click_policy="hide"
    
    # カラーバーの作成
    color_bar = ColorBar(color_mapper=mappers[color_target], location=(0,0), title=color_target)
    
    # プロットにレイアウト追加
    plot.add_layout(color_bar, 'right')
    plot.add_layout(plot.legend[0], "right")
    
    # インタラクティブウィジェットの設定
    @interact()
    def select_target(color_target=["target", "reliability"]):
        # 選択された色付け対象に応じて、グラフを更新
        if color_target == "target":
            r_train.glyph.fill_color = transform("target", mappers[color_target])
            if df_predict is not None:
                r_pred.glyph.fill_color = transform("prediction_label", mappers[color_target])
        else:
            r_train.glyph.fill_color = transform(color_target, mappers[color_target])
            if df_predict is not None:
                r_pred.glyph.fill_color = transform(color_target, mappers[color_target])
                
        color_bar.color_mapper=mappers[color_target]
        color_bar.title = color_target
        push_notebook()
    
    # プロットの表示
    show(plot, notebook_handle=True)
    push_notebook()    

    return plot
        